#
# Copyright 2008 TKK/ComNet
# Released under GPLv3. See LICENSE.txt for details.
# 

require 'rubygems'
require 'mongrel'
require 'logger'
require 'ostruct'
require 'fileutils'
require 'mongrel_extensions'
require 'resource'
require 'mhtml'
require 'cache'

module Blunder
  CONFIG = OpenStruct.new(
    :document_root  => nil,
    :cache_root     => nil,
    :resource_file  => nil,
    :index_files    => %w{index.html index.htm},
    :caching        => true,
    :port           => 3301,
    :eid            => nil,
    :debug          => false,
    :force_parser   => false,
    :clear_cache    => false,
    :log_file       => STDOUT,
    :log_level      => Logger::INFO
  )
        
  module Const
    include Mongrel::Const
    ERROR_500_RESPONSE = "HTTP/1.1 500 Internal Server Error\r\nConnection: close\r\n\r\nInternal Server Error"
  end

  # The handler is a subclass of DirHandler so that DirHandler can take care
  # of requests that didn't come from DTN.
  class BundleHandler < Mongrel::DirHandler
    def initialize(path)
      super(path, false)
      if CONFIG.clear_cache
        LOG.debug "Clearing cache (#{CONFIG.cache_root})"
        Cache.clear
      end
    end
    
    # Processes incoming requests. First checks if the request came from DTN
    # (i.e. current thread has [:bundle] set).
    # If it did, either serves a cached response or creates a new MHTML 
    # resource aggregate and returns that.
    # If not, lets DirHandler take care of it.
    def process(request, response)
      if Thread.current[:bundle]
        request_uri = request.params[Const::REQUEST_URI]
        LOG.debug "Parsed request: #{request_uri} (#{request.params['REMOTE_ADDR']})"
        response.start do |header, body|
          begin
            request_uri = Resource.normalize(request_uri)       
            if CONFIG.caching and mhtml = Cache.fetch(request_uri)
              LOG.debug "Serving response from cache"
            else
              resources = Resource.find(request_uri)
              LOG.debug "Found #{resources.length} associated resource(s)"
              resources = exclude_referer(resources, request.params) # FIXME: if resources.length = 1 we shouldnt do mhtml (?)
              mhtml, cache_metadata = Resource.create_mhtml(resources, request_uri)
              Cache.store(request_uri, mhtml, cache_metadata) if CONFIG.caching
            end
            header["Content-Type"] = "message/rfc822"
            body.write(mhtml)
          rescue ResourceNotLegal
            LOG.warn "Request outside docroot, sending no response"
            response.done = true
          rescue ResourceNotFound
            LOG.debug "Not found #{request_uri}, sending 404"
            response.socket.write(Const::ERROR_404_RESPONSE)
            response.done = true
          rescue
            LOG.warn "Unexpected error (#{$!}), sending 500"
            STDERR.puts $!.backtrace if LOG.level == Logger::DEBUG
            response.socket.write(Const::ERROR_500_RESPONSE)
            response.done = true
          end
        end
      else
        super(request, response) # Pass on to DirHandler
      end
    end

    # If the client sent a Referer header and it is from this domain, 
    # check what resources it yields and exclude those.
    def exclude_referer(resources, params)
      if resources.length > 0 and params['HTTP_REFERER'] and params['HTTP_REFERER'] =~ /#{Resource.base_location}/
        referer_path = URI.parse(params['HTTP_REFERER']).path
        already_sent = Resource.find(Resource.normalize(referer_path))        
        resources = resources - already_sent
        LOG.debug "Referer: #{referer_path} - sending #{resources.length} disjoint resource(s)"
      end
      resources
    end
    
  end
  
  # General extensions and shorthands
    
  String.class_eval do
    def /(o)
      File.join(self, o.to_s)
    end
  end
  
  Array.class_eval do
    def squash
      self.flatten.compact.uniq
    end
  end
  
  File.class_eval do
    class << self
      def last_modified(file)
        File.stat(file).mtime.httpdate
      end
    
      def etag(file)
        stat = File.stat(file)
        Const::ETAG_FORMAT % [stat.mtime.to_i, stat.size, stat.ino]
      end
    end
  end
  
end
