#
# Copyright 2008 TKK/ComNet
# Released under GPLv3. See LICENSE.txt for details.
#

module Blunder
  
  class CacheError < StandardError; end
  
  class Cache    
    class << self
      
      def full_path_to(file)
        File.expand_path(CONFIG.cache_root/file)
      end
      
      # Normalizes a requested path by appending it to the absolute cache root.
      # If the path is a directory, tries to find an index file.
      # Also makes sure that the path doesn't go outside the cache root.
      def normalize(request_uri)
        raise "Cache infringement" unless full_path_to(request_uri) =~ /#{CONFIG.cache_root}/
        if File.directory?(full_path_to(request_uri))
          request_uri += '/' if request_uri =~ /[^\/]$/
          request_uri = Cache.find_index(request_uri)
        end
        request_uri
      end   
      
      def meta(file)
        file + ".meta"
      end

      # Fetches a file from the cache. Returns the file data on success,
      # nil on failure.
      def fetch(file, check_staleness=true)
        path = full_path_to(normalize(file))
        if path.nil? or !File.exist? path
          return nil
        elsif check_staleness and stale? file
          LOG.debug "Cached resource was stale"
          return nil 
        end
        File.read(path)
      end

      # Stores a file to the cache. Creates directories if necessary.
      # By default, overwrites files in the cache. This can be changed
      # by setting the overwrite parameter to false.
      def store(file, data, metadata={}, overwrite=true)
        path = full_path_to(normalize(file))
        raise CacheError, "File already exists" if File.exist?(path) and overwrite == false
        FileUtils.makedirs(File.dirname(path))
        File.open(path, "w") { |f| f.write data }
        File.open(meta(path), "w") { |f| f.write metadata.to_yaml }
      rescue CacheError
        LOG.warn $!
      end    

      # Deletes everything from the cache. Set CONFIG.cache_root correctly or suffer.
      def clear(verbose=false)
        FileUtils.rm_rf(CONFIG.cache_root, :secure => true, :verbose => verbose)
        FileUtils.mkdir(CONFIG.cache_root, :verbose => verbose)
      end
      
      # Checks ETag and Expires from metadata to determine staleness.
      # Error condition (e.g. missing metadata) implies stale.
      def stale?(file)
        meta = YAML::load_file(meta(full_path_to(file)))
        etag = File.etag(Resource.full_path_to(file))
        if meta["ETag"] != etag or Time.parse(meta["Expires"]) < Time.now
          true
        else
          false
        end
      rescue
        true
      end

      # Tries to find an index file. 
      def find_index(path)
        CONFIG.index_files.each do |index|
          return path/index if File.exist? path/index
        end
        nil
      end
    end    
  end 
  
end
