#
# Copyright 2008 TKK/ComNet
# Released under GPLv3. See LICENSE.txt for details.
#

require 'dtn'

# These classes provide further abstractations over the Ruby bindings for
# the DTN2 library. The classes are heavily influenced by the similar DtnClient
# class by Dirk Kutscher <dku@tzi.org> (also the author of the Ruby bindings).
module Dtn
  SUCCESS = 0
  FAILURE = -1
  MAX_BUNDLE_MEM_PAYLOAD_SIZE = 50*1024
  
  # This is thrown when receiving bundles with a timeout value.
  class Timeout < RuntimeError; end  
  
  # Instances of the Client class represent client nodes connected to the
  # bundle router.
  class Client
    attr_reader :eid, :regid
    
    # Initializes the client: creates a new endpoint with the given URI,
    # opens a connection to the bundle router (@handle) and registers
    # the client.
    def initialize(uri="")
      @eid = Dtn_endpoint_id_t.new
      @eid.uri = uri
      @handle = Dtn::openHandle
      register
      raise "Error opening connection to bundle router" unless connected?
    end
    
    # Returns the endpoint id of the client, e.g. dtn://lpmbp.dtn/client.
    def uri
      @eid.uri
    end
    
    # Returns true when dtn_errno is 0.
    def connected?
      Dtn::dtn_errno(@handle) == 0
    end
    
    # Creates a DTN registration. The init_passive flag is set to 0, which means
    # that the client will be initially in active state. To change active/passive
    # state, bind and unbind can be used.
    def register
      r = Dtn_reg_info_t.new
      r.endpoint = @eid
      r.regid = 0
      r.failure_action = Dtn::DTN_REG_DEFER
      r.expiration = 10
      r.init_passive = 0
      r.script.script_len = 0
      r.script.script_val = ""
      res, @regid = Dtn::dtn_register(@handle, r)
      raise "Error registering DtnClient" unless res == SUCCESS
    end

    # Unregisters the client.
    def unregister
      raise "Unregistering the client failed" unless Dtn::dtn_unregister(@handle, @regid) == SUCCESS
    end
    
    # Closes the client connection to the bundle router.
    def close
      raise "Closing the client failed" unless Dtn::dtn_close(@handle) == SUCCESS
    end        
    
    # Sends the given bundle. Before sending, the source and replyto fields
    # of the bundle are populated with this client's endpoint id. Returns
    # the bundle id if successfully sent, nil if not.
    def send_bundle(bundle)
      bid = Dtn_bundle_id_t.new
      bundle.spec.source.uri = bundle.spec.replyto.uri = @eid.uri
      if Dtn::dtn_send(@handle, bundle.spec, bundle.payload, bid) == SUCCESS
        bid
      else
        nil
      end
    end
        
    # Receives a bundle. Creates a new Bundle instance on each call, so this
    # is not a good way to do this. We could use a instance variable
    # @recv_bundle for receiving, but then we would have to return a deep
    # copy of it. Object#clone won't do because that's just a shallow copy:
    # the referenced objects get overwritten with each call and bundle's
    # state will change while it's in the processing thread.
    # If a block is given, passes the received bundle to the block. If not,
    # returns the bundle.
    def receive_bundle(timeout, type=DTN_PAYLOAD_MEM)
      bundle = Bundle.new("", type)
      if recv(bundle.spec, bundle.payload, timeout) == SUCCESS
        if block_given?
          yield(bundle)
        else
          bundle
        end
      else
        bundle.payload.finalize if bundle.payload.respond_to? :finalize
        bundle.free
        raise Timeout
      end
    end

    # def find_registration
    # end    
    # 
    # def bind 
    #   Dtn::dtn_bind(@handle, @regid)
    # end
    # 
    # def unbind
    #   Dtn::dtn_unbind(@handle, @regid)
    # end        
    # 
    # def begin_poll(timeout)
    #   Dtn::dtn_begin_poll(@handle, timeout)
    # end
    # 
    # def cancel_poll
    #   Dtn::dtn_cancel_poll(@handle)
    # end
        
    private
    
    # Used internally for receiving bundles.
    def recv(spec, payload, timeout)
      Dtn::dtn_recv(@handle, spec, payload.location, payload, timeout)
    end
  end  
  
  
  # Represents a bundle. Contains the various bundle options (the "bundle spec")
  # and the payload, which can be one of three different types.
  class Bundle
    attr_accessor :spec, :payload
    
    # Creates a new bundle with the given destination URI. Sets the bundle
    # spec options and creates a new payload object of the given type.
    def initialize(destination_uri="", type=DTN_PAYLOAD_MEM)
      @spec = Dtn_bundle_spec_t.new
      @spec.dest.uri = destination_uri
      @spec.priority = COS_NORMAL
      @spec.dopts = DOPTS_NONE
      @spec.expiration = 60*10 # 60*60*60
      @spec.blocks.blocks_len = 0
      
      @payload = case type
      when DTN_PAYLOAD_MEM; MemPayload.new
      when DTN_PAYLOAD_FILE; FilePayload.new
      when DTN_PAYLOAD_TEMP_FILE; TempfilePayload.new
      else
        raise "Invalid payload type given"
      end      
    end
    
    # Frees the bundle payload.
    def free
      Dtn::dtn_free_payload(@payload)
    end    
  end
      
  
  # For bundles that are kept in memory.
  class MemPayload < Dtn_bundle_payload_t
    def initialize
      super
      self.location = DTN_PAYLOAD_MEM
      self.buf.buf_len = 0
      Dtn::initBundleMemPayload(self)
    end
    
    # Frees the memory allocated for the bundle. Must be called explicitly.
    def finalize
      Dtn::deleteBundleMemPayload(self)
    end
    
    # Sets the payload data.
    def data=(data)
      Dtn::setPayloadBufVal(self, data)
    end
    
    # Returns the payload data.
    def data
      Dtn::getPayloadBufVal(self, MAX_BUNDLE_MEM_PAYLOAD_SIZE)
    end
  end
  
  
  # For bundles that are kept in the file system.
  class FilePayload < Dtn_bundle_payload_t
    def initialize
      super
      self.location = DTN_PAYLOAD_FILE
      self.file = "/tmp/bundle_tmp_payload#{Time.now.usec+rand(100000)}"
    end
    
    # Sets the payload filename.
    def file=(name)
      self.filename.filename_val = name
      self.filename.filename_len = name.length
    end
    
    # Returns the payload filename.
    def file
      self.filename.filename_val
    end
    
    # Sets the payload data by writing into the payload file.
    def data=(data)
      File.open(self.file, 'w') { |f| f.write data }
    end
    
    # Returns the payload data by reading the payload file.
    def data
      File.read(self.file)
    end
  end
  
  
  # Equal to FilePayload, except uses a tempfile.
  class TempfilePayload < FilePayload # TODO: TempfilePayloads don't work methinks
    def initialize
      super
      self.location = DTN_PAYLOAD_TEMP_FILE
    end
  end
  
  
  # A faux socket. Defines the socket operations required by Mongrel's
  # HttpServer#process_client method, but uses internal IO buffers for
  # reading and writing. This is done so that we can use process_client
  # directly without having to monkey patch it.  
  class FauxSocket
    attr_reader :outdata
    
    def initialize(data)
      @in = StringIO.new(data)
      @out = StringIO.new
    end
        
    # Reads from the @in instance variable.
    def read
      @in.read
    end
    
    # Reads up to maxlen btyes from the @in instance variable.
    def readpartial(maxlen)
      if @in.respond_to? "readpartial"
        @in.readpartial(maxlen)
      else
        @in.read(maxlen)
      end
    end        
        
    # Writes to the @out instance variable.
    def write(string)
      @out.write(string)
    end
    
    # Returns the value of (@out.closed? and @in.closed?)
    def closed?
      @out.closed? and @in.closed?
    end
    
    # Writes the data in @out to the @outdata instance variable, from which
    # it can be obtained afterwards. Closes @in and @out IO buffers.
    def close
      @out.rewind
      @outdata = @out.read
      @in.close
      @out.close
    end
    
    def peeraddr
      [nil, nil, nil, nil]
    end    
  end
  

  class DtnSocket < FauxSocket
    attr_reader :bundle
    
    # A DtnSocket is initialized with a Bundle. The bundle's payload data
    # is put into the @in StringIO variable, which can be read as if it
    # were a socket. Similarly, the @out variable is created for writing.
    # After the DtnSocket is closed, the written data will be available in
    # the @outdata variable.
    def initialize(bundle)
      @bundle = bundle
      if @bundle.payload.location == DTN_PAYLOAD_MEM
        @in = StringIO.new(@bundle.payload.data)
        @out = StringIO.new
      else
        @in = File.open(@bundle.payload.file, "r")
        @out = StringIO.new
      end
      @outdata = String.new
    end
        
    # A convenience method for creating a response bundle directly from @bundle 
    # (@bundle.spec.replyto.uri is set as the destination uri) and @outdata.
    # Returns nil if @outdata is empty (HTTP parser failed or something else
    # went wrong), otherwise returns the bundle.
    def create_response_bundle(type=DTN_PAYLOAD_MEM)
      return nil if @outdata.empty?
      bundle = Bundle.new(@bundle.spec.replyto.uri, type)
      bundle.payload.data = @outdata
      bundle
    end    
    
    # IPSocket#peeraddr returns an array containing four elements:
    # "AF_INET", port number, host name, IP address. These don't really
    # make much sense in the bundle layer. This method will return 
    # nil, nil, source-eid, source-eid.
    def peeraddr
      [nil, nil, @bundle.spec.source.uri, @bundle.spec.source.uri]
    end
    
    # Find the HTTP request URI from a bundle payload
    def http_request_uri
      request_uri = @in.readline.match(/GET (.+) HTTP/)[1]
      @in.rewind
      request_uri
    rescue
      nil
    end
    
  end  
end
