#
# Copyright 2008 TKK/ComNet
# Released under GPLv3. See LICENSE.txt for details.
#

require 'mongrel'
require 'rmail'

module Mhtml
  MIME_TYPES = Mongrel::DirHandler::MIME_TYPES # TODO: remove dependency to mongrel?
  DEFAULT_CONTENT_TYPE = 'text/plain'

  # Represents an MHTML message. Basically just an RMail::Message with
  # some MHTML specific functionality.
  class Message < RMail::Message

    # Creates a new message and sets the required headers: Mime-Version, 
    # Content-Type and Content-Location (provided as an argument).
    def initialize(base_location)
      super()
      self.header.set('Mime-Version', '1.0')
      self.header.set('Content-Type', 'multipart/related')
      self.header.set('Content-Location', base_location)
    end

    # Overrides RMail::Message#add_part for convenience. Takes two arguments,
    # of which the first one can be a String or a Message object. The optional
    # second argument may contain a hash of MIME headers for the part.
    def add_part(part, headers=nil)
      if part.kind_of? String
        p = part
        part = RMail::Message.new
        part.body = p
      end      
      headers.each { |k,v| part.header.set(k, v) } if headers        
      super(part)
    end

    # Adds a part from the given filename. Sets the correct Content-Type and
    # Content-Location headers. By default, does not perform encoding on
    # binary files. This can be changed by setting the encode argument.
    # 
    def add_file_part(filename, content_location, headers={}, encode=false)
      return nil unless File.exist? filename
      file_data = File.read(filename)
      content_type = MIME_TYPES[/\.(\w)+$/.match(filename)[0]] || DEFAULT_CONTENT_TYPE
      headers['Content-Type'] = content_type
      headers['Content-Location'] = content_location.gsub(/^\//, '') # Remove leading slash
      if encode
        if content_type =~ /text\//
          file_data = RMail::Utils.quoted_printable_encode(file_data)
          headers['Content-Transfer-Encoding'] = 'quoted-printable'
        else
          file_data = RMail::Utils.base64_encode(file_data)
          headers['Content-Transfer-Encoding'] = 'base64'
        end
      end
      add_part(file_data, headers)
    end
    
    # Iterates through each part of the Mhtml structure. Decodes base64
    # encoded parts.
    def each_file_part
      @body.each do |part|
        if part.header['Content-Transfer-Encoding'] == 'base64'
          data = RMail::Utils.base64_decode(part.body)
        elsif part.header['Content-Transfer-Encoding'] == 'quoted-printable'
          data = RMail::Utils.quoted_printable_decode(part.body)
        else          
          data = part.body
        end        
        yield(part.header, data)
      end
    end
    
    # Parses a string containing an MHTML message into an Mhtml::Message object.
    def self.parse(data)
      rmail_message = RMail::Parser.read(data)
      mhtml_message = Mhtml::Message.new('')
      mhtml_message.header.replace(rmail_message.header.clone)
      mhtml_message.body = rmail_message.body
      mhtml_message
    end
        
  end  
end


# Extends/fixes a few things in the RMail library.
module RMail
  module Utils
    # Encodes the given string with base64
    def self.base64_encode(str)
      [str].pack("m")
    end
    # Encodes the given string with quoted-printable
    def self.quoted_printable_encode(str)
      [str].pack("M")
    end
  end

  class Header
    # Fix this method so that it doesn't create duplicate quotes 
    # in the header values (i.e. prevent this: type=""text/html"").
    def to_s
      to_string(true).gsub('""', '"')
    end
  end
end
