#
# Copyright 2008 TKK/ComNet
# Released under GPLv3. See LICENSE.txt for details.
#

require 'dtn_extensions'

module Mongrel
  
  # This class contains method definitions that override their counterparts
  # in HttpServer in order to implement bundle handling.
  class BundleHttpServer < HttpServer
    attr_accessor :dtn_client
    
    # Calls super() to set up the HTTP server and creates a DtnClient.
    def initialize(host, port, eid, num_processors=950, throttle=0, timeout=60)
      super(host, port, num_processors=950, throttle=0, timeout=60)
      @dtn_client = Dtn::Client.new(eid)
    end
        
    # Overrides HttpServer#run in Mongrel. Instead of running a single acceptor
    # thread to accept TCP connections, runs two threads: one for TCP and one
    # for DTN. Both acceptor threads spawn new worker threads on incoming
    # requests.
    def run
      run_dtn
      super
      [@acceptor, @dtn_acceptor]
    end
    
    # Runs the main loop for receiving bundled requests through DTN.
    def run_dtn
      @dtn_acceptor = Thread.new do
        loop do
          begin
            bundle = @dtn_client.receive_bundle(1, Dtn::DTN_PAYLOAD_FILE)
            thread = Thread.new(bundle) do |b| 
              # This is used in Handler#process to determine if the request 
              # came from DTN or plain TCP
              Thread.current[:bundle] = true
              process_bundle(b)
            end
            # Prevents concurrency! (lower priority threads will never run)
            thread.priority = 1 # Set higher priority for the worker thread
          rescue Dtn::Timeout
            Thread.pass
            # This might help with concurrency (by allowing worker threads
            # to be scheduled when the listener is sleeping). Comment the
            # priority line above if you use this.
            #sleep(0.1)
          rescue
            STDERR.puts "Unhandled exception in DTN listen loop: #{$!}"
          end
        end
      end
      @dtn_acceptor
    end
        
    # Processes a bundle which contains an HTTP request. First creates a
    # DtnSocket from the bundle and calls HttpServer#process_client, which
    # will process the HTTP request and write the response to the DtnSocket.
    # Creates a new bundle, sets the response as the payload, and sends the 
    # bundle.
    def process_bundle(bundle)
      dtn_socket = Dtn::DtnSocket.new(bundle)
      process_client(dtn_socket)
      response_bundle = dtn_socket.create_response_bundle(Dtn::DTN_PAYLOAD_FILE)
      if response_bundle # If nil, just don't respond. Most likely HTTP parser failed. 
        bid = @dtn_client.send_bundle(response_bundle)
        raise "Sending the response bundle failed" if bid.nil?
      end
    end
    
    # Raw HTTP bundling, lifted from version 0.5. If someone wants to enable it,
    # this is a good starting point.
    #
    # Processes a bundle. First extracts the request URI from the HTTP request
    # and finds associated resources. For each resource, invokes process_client
    # and captures the response. Then, concatenates the responses and sends
    # them back in a single bundle.
    # def process_bundle_raw_http(bundle)
    #   dtn_socket = Dtn::DtnSocket.new(bundle)
    #   resources = Resource.find(dtn_socket.http_request_uri)
    #   response_data = ""      
    #   process_client(dtn_socket)
    #   response_data << dtn_socket.outdata + "\r\n\r\n"
    #   resources.each do |r|
    #     r = '/'+r unless r =~ /^\// # Add leading slash if missing
    #     faux_socket = Dtn::FauxSocket.new("GET #{r} HTTP/1.1\r\n\r\n")
    #     process_client(faux_socket)
    #     response_data << faux_socket.outdata + "\r\n\r\n"
    #   end
    #   response_bundle = Dtn::Bundle.new(bundle.spec.replyto.uri, Dtn::DTN_PAYLOAD_FILE)
    #   response_bundle.payload.data = response_data
    #   bid = @dtn_client.send_bundle(response_bundle)
    #   raise "Sending the response bundle failed" if bid.nil?
    # rescue
    #   STDERR.puts "Unhandled exception #{$!}"
    # end
        
    # Shuts down the HTTP server, then unregisters and closes the Dtn::Client.
    def graceful_shutdown
      super
      @dtn_client.unregister
      @dtn_client.close
    rescue
      STDERR.puts $!
    end
  end
end
