#
# Copyright 2008 TKK/ComNet
# Released under GPLv3. See LICENSE.txt for details.
#

require 'find'

module Blunder
  
  class ResourceNotFound < StandardError; end
  class ResourceNotLegal < StandardError; end
  
  class Resource
    class << self
      # "Class level accessors"
      attr_accessor :resource_yaml, :base_location
      
      def full_path_to(request_uri)
        File.expand_path(CONFIG.document_root/request_uri)
      end
      
      
      # Preprocess the request_uri: check that the request is legal and
      # the requested file exists. If the request is for a directory, try
      # to find the index file.
      def normalize(request_uri)
        raise ResourceNotLegal unless Resource.legal? request_uri
        raise ResourceNotFound unless Resource.exist? request_uri
        if File.directory?(full_path_to(request_uri))
          request_uri += '/' if request_uri =~ /[^\/]$/ # Append trailing slash if missing
          request_uri = Resource.find_index(request_uri)
        end
        request_uri
      end
            
      # Finds the associated resources (files) for the given request path. 
      # Returns an array of filenames. The nofollow parameter is only used 
      # internally to break circular references in recursion.
      # If @resource_yaml is nil, tries to find the associated resources
      # by parsing the requested (HTML) file.
      def find(request_path, nofollow=request_path)
        return parse_html(request_path) if @resource_yaml.nil?
        resources = []
        matched = []

        # Get resources with * only on first iteration
        resources += @resource_yaml['*'] if nofollow == request_path and @resource_yaml['*']                
        resources += @resource_yaml[request_path] if @resource_yaml[request_path]
              
        resources.each do |r|
          if r.kind_of?(Hash) and r.has_key?('also')            
            next if r['also'] == nofollow # Break circular reference
            matched << Resource.get(r['also']) # Include the also: resource
            matched << Resource.find(r['also'], request_path)
          else
            matched << Resource.get(r)
          end
        end
        matched.squash
      end
      
      # Tries to parse the resources from the requested file (obviously only 
      # works for HTML files). Searches for <link>, <script>, <img> and <style>
      # elements and gets their contents, unless it is an external reference 
      # (http://xyz, where xyz is NOT the domain in Resource.base_location).
      # Only works for well-formed documents with quoted attribute values.
      # Works for both absolute and relative paths.
      def parse_html(request_path, include_favicon=false)
        return [] unless request_path =~ /\.(html|htm)$/      # Only parse html documents
        base = Regexp.escape(Resource.base_location)          # Server's own domain, http://something
        chars = /[\w\.\/-]/                                   # Allowed characters in a resource name
        patterns = [
          /<link.*?href="(?:#{base})?(#{chars}*)"/i,          # Stylesheets and other links
          /@import.*?[url]?\(?['|"]?(#{chars}*)['|"]?\)?;/i,  # Imported stylesheets
          /<script.*?src="(?:#{base})?(#{chars}*)"/i,         # Javascripts
          /<img.*?src="(?:#{base})?(#{chars}*)"/i             # Images
        ]
        file_content = File.read(full_path_to(request_path))
        resources = file_content.scan(Regexp.union(*patterns)).squash
        
        # Expand relative paths (i.e. change relative file references that appear
        # somewhere deep in the directory tree to absolute ones)
        resources.map! { |r| r =~ /^\// ? r : File.dirname(request_path)/r }
        
        # Parse all css files
        css_resources = resources.select { |r| r =~ /\.css$/ }
        css_resources.each { |r| resources << parse_css(r) if Resource.exist?(r) }

        # Browsers always ask for this file
        resources << "/favicon.ico" if include_favicon
        resources.squash
      end
      
      # Parses resources from a CSS file. Searches for url("filename.ext") and
      # @import "filename.ext". By default, searches recursively (this CSS file
      # may refer to another CSS file, which may refer to another CSS file...).
      # Watch out for circular references (A @imports B, B @imports A)!
      def parse_css(request_path, recursive=true)
        chars = /[\w\.\/-]/
        patterns = [
          /url\(['|"]?(#{chars}*)['|"]?\)/i,                  # Url references
          /@import.*?[url]?\(?['|"]?(#{chars}*)['|"]?\)?;/i   # Imported stylesheets
        ]
        file_content = File.read(full_path_to(request_path))
        resources = file_content.scan(Regexp.union(*patterns)).squash
        resources.map! { |r| r =~ /^\// ? r : File.dirname(request_path)/r }

        # Parse files referred to with @import ad infinitum
        if recursive
          css_resources = resources.select { |r| r =~ /\.css$/ }
          css_resources.each { |r| resources << parse_css(r) if Resource.exist?(r) }      
        end
        resources.squash
      end
            
      # Get the file(s) represented by request_string (which may contain
      # a wildcard, *).
      def get(request_string)
        if request_string =~ /\/\*$/ # Wildcard: get all files recursively
          docroot = File.expand_path(CONFIG.document_root)
          files = []
          Find.find(full_path_to(File.dirname(request_string))) do |path|
            next if File.directory? path
            next if File.basename(path) =~ /^\./ # Ignore files starting with dot
            files << path.sub(docroot, '') # Get the relevant part of path
          end
          files
        else # A single file
          File.exist?(full_path_to(request_string)) ? request_string : nil
        end
      end
      
      # Creates an MHTML message from the resources that match the request URI.
      # The top level Content-Location is the host part of the request URI, e.g.
      # http://lpmbp.dtn/. The parts have a Content-Location relative to this, e.g.
      # index.html or images/logo.png (leading slash is allowed but not required).
      # SHOULD NOT trigger 404 if an associated resource is not found. Should just
      # silently skip those.
      def create_mhtml(resources, request_uri)
        mhtml = Mhtml::Message.new(@base_location)
        cache_metadata = {}
        
        resources.unshift(request_uri) # Push the requested file into the array of resources
        max_age = 60*60*24*30 # TODO: Something else than fixed expiry after 30 days
        resources.each do |filename|
          if File.exist? full_path_to(filename)
            expires = (Time.now + max_age).httpdate
            etag = File.etag(full_path_to(filename))
            mhtml.add_file_part(full_path_to(filename), filename, {
              'X-Expires' => expires,
              'X-Cache-Control' => "max-age=#{max_age}",
              'X-ETag' => etag,
              'X-Last-Modified' => File.last_modified(full_path_to(filename))
            })
            if filename == request_uri
              cache_metadata['Expires'] = expires
              cache_metadata['ETag'] = etag
            end
          end
        end
        [mhtml.to_s, cache_metadata]
      end            
      
      protected
      
      # Checks if the requested resource exists.
      def exist?(request_uri)                
        if File.directory? full_path_to(request_uri)
          !Resource.find_index(request_uri).nil?
        else
          File.exist? full_path_to(request_uri)
        end
      end
      
      # Ensures that the requested resource is within the resource path (docroot).
      # Does this by checking that the final expanded path to the file contains
      # the docroot.
      def legal?(request_uri)
        full_path_to(request_uri) =~ /#{CONFIG.document_root}/ ? true : false
      end
            
      # Tries to find an index file. Returns the filename or raises ResourceNotFound.
      def find_index(request_uri)
        CONFIG.index_files.each do |index|
          return request_uri/index if File.exists? full_path_to(request_uri/index)
        end
        raise ResourceNotFound
      end
      
    end # class << self
  end
  
end
