#
# Copyright 2008 TKK/ComNet
# Released under GPLv3. See LICENSE.txt for details.
#

require File.dirname(__FILE__) + '/spec_helper'

describe Resource, "when resource dependency file is available" do
  
  before(:all) do
    Resource.resource_yaml = YAML::load_file(CONFIG.resource_file)
  end
  
  it "should return an array of 2 resources for non-matching request string" do
    r = Resource.find('/non/existent')
    r.should be_an_instance_of(Array)
    r.should have(2).elements
  end
  
  it "should return an array of 4 resources for /" do
    r = Resource.find('/')
    r.should be_an_instance_of(Array)
    r.should have(4).elements
  end
  
  it "should break a circular reference (A has 'also: B' and B has 'also: A')" do
    lambda { Resource.find('/test.html') }.should_not raise_error
  end
  
  it "should not return duplicate resources" do
    r = Resource.find('/test.html')
    r.should have(5).elements
    r.should == r.uniq
  end
  
  it "should return all files recursively when using a wildcard" do
    r = Resource.find('/recursive.html')
    r.should have(6).elements
    r.should include('/dir/dir2/dir2_test.jpg')
  end
  
  it "wildcard should not include files whose name start with dot (.)" do
    r = Resource.find('/recursive.html')
    r.should_not include('/dir/.ignore_me')
  end
end


describe Resource, "when resource dependency file is NOT available" do
  
  before(:all) do
    Resource.base_location = 'http://lpmbp.dtn'
    Resource.resource_yaml = nil
  end
    
  it "should return an empty array if the requested file is not HTML" do
    r = Resource.find('/test.jpg')
    r.should be_an_instance_of(Array)
    r.should be_empty
  end

  # Note: this spec must know whether include_favicon is set in Resource.parse_html
  it "should find external CSS, Javascript and image files from an HTML document" do
    r = Resource.find('/test.html', false)
    r.should have(5).elements
    r.sort.should == %w{/images/DSC_0039.jpg /images/DSC_0064.jpg /imported.css /script.js /style.css}
  end

  # Actually, assume valid (X)HTML. If we want to parse broken documents, we can do
  # that later (maybe with hpricot)
  # it "should also parse resources out of more challenging HTML documents" do
  #   html_snippets = %w{
  #     <link type=text/css href=styles.css>
  #     <style type="text/css" media="all">@import "/css/styles.css"</style>
  #   }
  # end
end


describe Resource, "when creating MTHML" do
  
  before(:all) do
    Resource.base_location = 'http://lpmbp.dtn'
    Resource.resource_yaml = YAML::load(<<-YAML
      /test.html:
        - test.jpg
    YAML
    )
    request_path = '/test.html'
    resources = Resource.find('/test.html')
    @mhtml, @metadata = Resource.create_mhtml(resources, request_path)
    @message = Mhtml::Message.parse(@mhtml)
  end
  
  
  it "should produce a string that can be parsed into an Mhtml::Message object" do
    @message.should be_an_instance_of(Mhtml::Message)
  end  
  
  it "should include the requested file and its dependencies" do
    locations = %w{test.html test.jpg}.map { |file| File.join(Resource.base_location, file) }
    @message.each_file_part do |header, data|
      location = File.join(Resource.base_location, header['Content-Location'])
      locations.should include(location)
    end
  end
  
  it "should include ETag and Expires metadata" do
    @metadata.should include('ETag')
    @metadata.should include('Expires')
  end
end


describe Resource, "when accessing resources" do
  
  before(:all) do
    Resource.resource_yaml = YAML::load_file(CONFIG.resource_file)
    Resource.base_location = 'http://lpmbp.dtn'
  end
  
  it "should prevent retrieving files from outside the docroot" do
    illegal_paths = [
      '../file.ext',
      '../../file.ext',
      './../file.ext',
      '//../file.ext',
      './///////.././file.ext'
    ]
    illegal_paths.each do |p| 
      lambda { Resource.normalize(p) }.should raise_error(ResourceNotLegal)
    end
  end
  
  it "should fail if the requested resource doesn't exist" do
    lambda { Resource.normalize('/non/existent/file.html') }.should raise_error(ResourceNotFound)
  end
  
  it "should find the index file when requesting for /" do
    Resource.normalize('/').should == '/index.html'
  end
  
  it "should raise error if requesting for a directory and there is no index file" do
    lambda { Resource.normalize('/dir/') }.should raise_error(ResourceNotFound)
  end
    
end
