#
# Copyright 2008 TKK/ComNet
# Released under GPLv3. See LICENSE.txt for details.
#

require 'rubygems'
require 'mongrel'
require 'logger'
require 'timeout'
require 'dtn'
require 'fileutils'
require 'mhtml'
require 'dtn_extensions'
require 'cache'

module Mongrel
  module Const
    ERROR_504_RESPONSE = "HTTP/1.1 504 Gateway Time-out\r\nConnection: close\r\n\r\nGateway Time-out"
    METADATA_HEADERS = %w{Expires ETag Cache-Control Last-Modified}
  end
  
  class HttpParams    
    # Reconstructs an HTTP header from the params hash that Mongrel produces.
    # Basically does this: HTTP_REQUEST_METHOD['GET'] -> "Request-Method: GET\r\n"
    def to_http_header
      header = "#{self['REQUEST_METHOD']} #{self['REQUEST_PATH']} #{self['HTTP_VERSION']}\r\n"
      self.each_pair do |k,v|
        if k =~ /^HTTP_(?!VERSION)/ # Match HTTP_* but not HTTP_VERSION
          header << k.downcase.gsub(/^http_/, '').split('_').map{|p| p.capitalize}.join('-')
          header << ": #{v}\r\n"
        end
      end
      header << "\r\n"
    end
  end
  
  class BundleProxy < HttpServer
    attr_reader :dtn_client
    
    def initialize(host, port, eid, num_processors=950, throttle=0, timeout=60)
      super(host, port, num_processors=950, throttle=0, timeout=60)
      @dtn_client = Dtn::Client.new(eid)
    end
    
    # Shuts down the HTTP server, then unregisters and closes the Dtn::Client.
    def graceful_shutdown
      super
      @dtn_client.unregister
      @dtn_client.close
    rescue
      STDERR.puts $!
    end
    
    # Kills the worker threads and shuts down.
    def hard_shutdown
      STDERR.puts "Killing #{@workers.list.length} worker threads"
      @workers.list.each { |worker| worker.kill }
      @dtn_client.unregister
      @dtn_client.close
    rescue
      STDERR.puts $!      
    end
  end
end

class Handler < Mongrel::HttpHandler
  
  def initialize
    Cache.clear(true) if CONFIG.clear_cache
  end

  # Processes incoming HTTP requests (from the browser). First checks the
  # local cache to see if the requested resource is available. If not, 
  # puts the request in a bundle and sends it over DTN to the same request URI.
  # For example, if the HTTP request was for http://lpmbp.dtn/index.html, the
  # bundle would be sent to dtn://lpmbp.dtn. After sending the
  # bundle, the handler begins waiting for bundles with a timeout period.
  # If a bundle is received, the contents are cached and the handler checks
  # if the received resource was the one that the client requested. If it was,
  # it is served to the client. If not, the handler keeps waiting for bundles
  # until the timeout period expires, at which point 504 is returned.
  
  # TODO: Send bundles from random EID (create a new Client instance on each send)
  def process(request, response)
    response.start do |head, body|
      request_uri = request.params['REQUEST_URI']
      LOG.info "Request: #{request_uri}"
      cached = Cache.fetch(request_uri)
      if cached
        LOG.info "Cache hit: #{request_uri}"
        body.write(cached)
      else
        LOG.info "Cache miss: #{request_uri} -- sending bundled request to #{request.params['HTTP_HOST']}"
        bundle = Dtn::Bundle.new("dtn://" + request.params['HTTP_HOST'])
        bundle.payload.data = request.params.to_http_header + request.body.read
        listener.dtn_client.send_bundle(bundle)
        Timeout::timeout(60*5) do # Wait this many seconds for response before returning 504
          begin
            listener.dtn_client.receive_bundle(1, Dtn::DTN_PAYLOAD_FILE) do |bundle|
              mhtml = Mhtml::Message.parse(bundle.payload.data)
              # Skip if we didn't get 200 OK
              next unless mhtml.multipart? and mhtml.header[0] =~ /HTTP\/1\.\w 200/
              mhtml.each_file_part do |header, data|
                base_location = mhtml.header['Content-Location']
                part_location = header['Content-Location']
                location = File.join(base_location, part_location)
                metadata = {}
                header.each do |field, value|
                  field = field.sub(/^X-/, '') # Remove leading X-
                  metadata[field] = value if Mongrel::Const::METADATA_HEADERS.include? field
                end
                Cache.store(location, data, metadata)
                LOG.info "Cached: #{part_location}"              
              end
              cached = Cache.fetch(request_uri)
              body.write(cached) if cached
            end
          rescue Dtn::Timeout
            Thread.pass
            retry
          end
        end # Timeout::timeout
      end
    end
  rescue Timeout::Error
    LOG.warn "Timeout"
    response.socket.write(Mongrel::Const::ERROR_504_RESPONSE)
    response.done = true
  end   
end
